/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.web.beans.view;

import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.faces.application.Application;
import javax.faces.application.FacesMessage;
import javax.faces.component.UIComponent;
import javax.faces.component.html.HtmlOutputText;
import javax.faces.context.FacesContext;

import kieker.analysis.AnalysisController;
import kieker.analysis.model.analysisMetaModel.MIDisplayConnector;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.MIView;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.monitoring.core.registry.Registry;
import kieker.webgui.common.exception.DisplayNotFoundException;
import kieker.webgui.common.exception.InvalidAnalysisStateException;
import kieker.webgui.common.exception.ProjectLoadException;
import kieker.webgui.domain.DisplayType;
import kieker.webgui.service.IProjectService;
import kieker.webgui.web.beans.application.GlobalPropertiesBean;
import kieker.webgui.web.beans.application.ProjectsBean;
import kieker.webgui.web.utility.CockpitLayout;

import org.primefaces.component.chart.line.LineChart;
import org.primefaces.component.chart.metergauge.MeterGaugeChart;
import org.primefaces.component.dashboard.Dashboard;
import org.primefaces.component.panel.Panel;
import org.primefaces.component.selectcheckboxmenu.SelectCheckboxMenu;
import org.primefaces.model.DashboardColumn;
import org.primefaces.model.DashboardModel;
import org.primefaces.model.DefaultDashboardColumn;
import org.primefaces.model.DefaultDashboardModel;
import org.primefaces.model.chart.CartesianChartModel;
import org.primefaces.model.chart.LineChartSeries;
import org.primefaces.model.chart.MeterGaugeChartModel;

import net.vidageek.mirror.dsl.Mirror;
import net.vidageek.mirror.exception.MirrorException;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

/**
 * This class is a {@code Spring} managed bean containing the necessary data behind an instance of the cockpit.<br>
 * <br/>
 * 
 * It has view scope to make sure that one user (even in one session) can open multiple projects at a time without causing any problems.
 * 
 * @author Nils Christian Ehmke
 */
@Component
@Scope("view")
public class CurrentCockpitBean {

	private static final Log LOG = LogFactory.getLog(CurrentCockpitBean.class);

	private static final String DASHBOARD_ID = "dashboard";
	private static final String PANEL_ID_PREFIX = "panel_";
	private static final int NUMBER_OF_COCKPIT_COLUMNS = 2;

	private final Registry<MIDisplayConnector> displayConnectors = new Registry<MIDisplayConnector>();

	private CockpitLayout cockpitLayout;
	private DashboardModel dashboardModel;
	private Dashboard dashboard;
	private String projectName;
	private MIProject project;
	private MIView activeView;

	@Autowired
	private GlobalPropertiesBean globalPropertiesBean;
	@Autowired
	private IProjectService projectService;
	@Autowired
	private ProjectsBean projectsBean;

	/**
	 * Creates a new instance of this class. <b>Do not call this constructor manually. It will only be accessed by Spring.</b>
	 */
	public CurrentCockpitBean() {
		// No code necessary
	}

	/**
	 * This method initializes the bean by using the current project name to load the project. <b>Do not call this method manually. It will only be accessed by
	 * Spring.</b>
	 */
	public void initalize() {
		try {
			// Make sure that the initialization will only be done for the init request.
			if (!FacesContext.getCurrentInstance().isPostback()) {
				this.loadProject();

				if (this.project != null) {
					this.createDashboardComponent();
					this.loadCockpitLayout();
				}
			}
		} catch (final ProjectLoadException ex) {
			CurrentCockpitBean.LOG.error("An error occured while loading the project.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, this.globalPropertiesBean.getMsgProjectLoadingException());
		} catch (final NullPointerException ex) {
			// This exception occurs, when the projectsBean has not been initialized
			CurrentCockpitBean.LOG.error("An error occured while loading the project.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, this.globalPropertiesBean.getMsgProjectLoadingException());
		}
	}

	/**
	 * Sets the active view to a new value and updates the dashboard.
	 * 
	 * @param activeView
	 *            The new active view.
	 */
	public void setActiveView(final MIView activeView) {
		this.activeView = activeView;

		this.fillDashboardWithActiveView();
	}

	/**
	 * Sets the dashboard of this bean.
	 * 
	 * @param dashboard
	 *            The new dashboard.
	 */
	public void setDashboard(final Dashboard dashboard) {
		this.dashboard = dashboard;

		// This step is necessary as we lose otherwise the dashboard model
		this.dashboard.setModel(this.dashboardModel);
	}

	public MIView getActiveView() {
		return this.activeView;
	}

	public Dashboard getDashboard() {
		return this.dashboard;
	}

	public MIProject getProject() {
		return this.project;
	}

	public void setProjectName(final String newName) {
		this.projectName = newName;
	}

	public String getProjectName() {
		return this.projectName;
	}

	/**
	 * Checks whether the analysis is currently running.
	 * 
	 * @return true if and only if the analysis is running.
	 */
	public boolean isAnalysisRunning() {
		try {
			return this.projectService.getCurrentState(this.projectName) == AnalysisController.STATE.RUNNING;
		} catch (final NullPointerException ex) {
			// This exception can occur, when the projectsBean has not been initialized
			LOG.warn("A null pointer exception occured.", ex);
		}
		return false;
	}

	/**
	 * Checks whether the analysis is currently in the ready state.
	 * 
	 * @return true if and only if the analysis is ready to be started.
	 */
	public boolean isAnalysisReady() {
		try {
			return this.projectService.getCurrentState(this.projectName) == AnalysisController.STATE.READY;
		} catch (final NullPointerException ex) {
			// This exception can occur, when the projectsBean has not been initialized
			LOG.warn("A null pointer exception occured.", ex);
		}
		return false;
	}

	/**
	 * Checks whether the analysis is not available.
	 * 
	 * @return true if and only if the analysis is <b>not</b> available.
	 */
	public boolean isAnalysisNotAvailable() {
		try {
			return this.projectService.getCurrentState(this.projectName) == null;
		} catch (final NullPointerException ex) {
			// This exception can occur, when the projectsBean has not been initialized
			LOG.warn("A null pointer exception occured.", ex);
		}
		return true;
	}

	/**
	 * Checks whether the analysis is currently terminated.
	 * 
	 * @return true if and only if the analysis has been terminated.
	 */
	public boolean isAnalysisTerminated() {
		try {
			return this.projectService.getCurrentState(this.projectName) == AnalysisController.STATE.TERMINATED;
		} catch (final NullPointerException ex) {
			// This exception can occur, when the projectsBean has not been initialized
			LOG.warn("A null pointer exception occured.", ex);
		}
		return false;
	}

	/**
	 * Checks whether the analysis is currently in the failed state.
	 * 
	 * @return true if and only if the analysis has failed.
	 */
	public boolean isAnalysisFailed() {
		try {
			return this.projectService.getCurrentState(this.projectName) == AnalysisController.STATE.FAILED;
		} catch (final NullPointerException ex) {
			// This exception can occur, when the projectsBean has not been initialized
			LOG.warn("A null pointer exception occured.", ex);
		}
		return false;
	}

	/**
	 * Updates all displays of the currently active view.
	 */
	public void updateDisplaysOfActiveView() {
		if (this.projectName != null) {
			for (final UIComponent component : this.dashboard.getChildren()) {
				// Get the correct display type of the current component
				final int componentID = Integer.valueOf(component.getId().replace(PANEL_ID_PREFIX, ""));
				final MIDisplayConnector displayConnector = this.displayConnectors.get(componentID);

				if (this.projectService.getCurrentState(this.projectName) != null) {
					final DisplayType displayType = this.projectService.getDisplayType(this.projectName, this.activeView.getName(), displayConnector.getName());

					if (displayType != null) {
						// Update the element depending on the display type
						switch (displayType) {
						case PLAIN_TEXT:
							this.updatePlainTextDisplay(component, displayConnector.getName());
							break;
						case HTML_TEXT:
							this.updateHTMLTextDisplay(component, displayConnector.getName());
							break;
						case IMAGE:
							this.updateImageDisplay(component, displayConnector.getName());
							break;
						case XY_PLOT:
							this.updateXYPlotDisplay(component, displayConnector.getName());
							break;
						case METER_GAUGE:
							this.updateMeterGaugeDisplay(component, displayConnector.getName());
							break;
						default:
							// Unknown type
							this.updateDisplayWithDefaultContent(component);
							break;
						}
					} else {
						this.updateDisplayWithDefaultContent(component);
					}
				} else {
					// If the analysis has not yet been initialized, we add a simple note to each of the panels
					this.updateDisplayWithDefaultContent(component);
				}
			}
		}
	}

	private void loadCockpitLayout() {
		this.cockpitLayout = new CockpitLayout(this.project, this.projectService.getCockpitLayout(this.projectName), NUMBER_OF_COCKPIT_COLUMNS);
	}

	private void createDashboardComponent() {
		final FacesContext facesContext = FacesContext.getCurrentInstance();
		final Application application = facesContext.getApplication();

		// Create the Primefaces dashboard component
		this.dashboard = (Dashboard) application.createComponent(facesContext, "org.primefaces.component.Dashboard", "org.primefaces.component.DashboardRenderer");
		this.dashboard.setId(DASHBOARD_ID);

		// Create the corresponding model with the correct number of columns
		this.dashboard.setModel(new DefaultDashboardModel());

		for (int i = 0; i < NUMBER_OF_COCKPIT_COLUMNS; i++) {
			this.dashboard.getModel().addColumn(new DefaultDashboardColumn());
		}

		// Remember the dashboard model
		this.dashboardModel = this.dashboard.getModel();
	}

	private void loadProject() throws ProjectLoadException {
		this.project = this.projectsBean.openProject(this.projectName);
	}

	private void fillDashboardWithActiveView() {
		// Dump the old entries
		this.clearDashboard();

		// Now add the entries from the active view
		if (this.activeView != null) {
			final List<List<MIDisplayConnector>> layout = this.cockpitLayout.getCurrentLayout(this.activeView);

			for (int columnIndex = 0; columnIndex < NUMBER_OF_COCKPIT_COLUMNS; columnIndex++) {
				final DashboardColumn column = this.dashboard.getModel().getColumn(columnIndex);

				for (final MIDisplayConnector displayConnector : layout.get(columnIndex)) {
					final Panel panel = this.createPanelFromDisplayConnector(displayConnector);

					// Add the panel to the dashboard and to its model
					this.dashboard.getChildren().add(panel);
					column.addWidget(panel.getId());
				}
			}
		}
	}

	private void clearDashboard() {
		// Run through all columns of the dashboard and remove the items
		final List<DashboardColumn> columns = this.dashboard.getModel().getColumns();
		for (final DashboardColumn column : columns) {
			column.getWidgets().clear();
		}

		// Now clear the dashboard itself
		this.dashboard.getChildren().clear();
	}

	private Panel createPanelFromDisplayConnector(final MIDisplayConnector connector) {
		final FacesContext facesContext = FacesContext.getCurrentInstance();
		final Application application = facesContext.getApplication();

		// Create the Primefaces panel component
		final Panel panel = (Panel) application.createComponent(facesContext, "org.primefaces.component.Panel", "org.primefaces.component.PanelRenderer");

		// Set the usual properties of the panel
		panel.setId(this.displayConnectorToID(connector));
		panel.setHeader(connector.getName());
		panel.setClosable(false);
		panel.setToggleable(false);

		return panel;
	}

	private String displayConnectorToID(final MIDisplayConnector displayConnector) {
		return PANEL_ID_PREFIX + this.displayConnectors.get(displayConnector);
	}

	private void updateDisplayWithDefaultContent(final UIComponent component) {
		final boolean isAlreadyCorrectlyInitialized = (component.getChildCount() == 1) && (component.getChildren().get(0) instanceof HtmlOutputText);
		final HtmlOutputText htmlOutputText;

		// Avoid unnecessary component creation. Use the existing components if they are already existing and from the correct type.
		if (isAlreadyCorrectlyInitialized) {
			htmlOutputText = (HtmlOutputText) component.getChildren().get(0);
		} else {
			htmlOutputText = new HtmlOutputText();

			component.getChildren().clear();
			component.getChildren().add(htmlOutputText);
		}

		htmlOutputText.setValue("N/A");
	}

	private void updateHTMLTextDisplay(final UIComponent component, final String displayConnectorName) { // NOPMD (Not implemented yet)
		// Not implemented yet
		this.updateDisplayWithDefaultContent(component);
	}

	private void updateImageDisplay(final UIComponent component, final String displayConnectorName) { // NOPMD (Not implemented yet)
		// Not implemented yet
		this.updateDisplayWithDefaultContent(component);
	}

	private void updatePlainTextDisplay(final UIComponent component, final String displayConnectorName) {
		final boolean isAlreadyCorrectlyInitialized = (component.getChildCount() == 1) && (component.getChildren().get(0) instanceof HtmlOutputText);
		final HtmlOutputText htmlOutputText;

		// Avoid unnecessary component creation. Use the existing components if they are already existing and from the correct type.
		if (isAlreadyCorrectlyInitialized) {
			htmlOutputText = (HtmlOutputText) component.getChildren().get(0);
		} else {
			htmlOutputText = new HtmlOutputText();

			component.getChildren().clear();
			component.getChildren().add(htmlOutputText);
		}

		try {
			final Object displayObject = this.projectService.getDisplay(this.projectName, this.activeView.getName(), displayConnectorName);
			final String text = (String) new Mirror().on(displayObject).invoke().method("getText").withoutArgs();

			htmlOutputText.setValue(text);

		} catch (final DisplayNotFoundException ex) {
			CurrentCockpitBean.LOG.warn("Display not found.", ex);
			this.updateDisplayWithDefaultContent(component);
		} catch (final MirrorException ex) {
			CurrentCockpitBean.LOG.warn("Reflection exception.", ex);
			this.updateDisplayWithDefaultContent(component);
		} catch (final InvalidAnalysisStateException ex) {
			CurrentCockpitBean.LOG.info("Project is in invalid state.", ex);
			this.updateDisplayWithDefaultContent(component);
		}
	}

	@SuppressWarnings("unchecked")
	private void updateMeterGaugeDisplay(final UIComponent component, final String displayConnectorName) {
		final boolean isAlreadyCorrectlyInitialized = (component.getChildCount() == 1) && (component.getChildren().get(0) instanceof MeterGaugeChart);

		final MeterGaugeChart meterGaugeChart;
		final MeterGaugeChartModel meterGaugeChartModel;

		// Avoid unnecessary component creation. Use the existing components if they are already existing and from the correct type.
		if (isAlreadyCorrectlyInitialized) {
			meterGaugeChart = (MeterGaugeChart) component.getChildren().get(0);
			meterGaugeChartModel = meterGaugeChart.getValue();
		} else {
			final FacesContext facesContext = FacesContext.getCurrentInstance();
			final Application application = facesContext.getApplication();

			// Create the Primefaces chart component
			meterGaugeChart = (MeterGaugeChart) application.createComponent(facesContext, "org.primefaces.component.chart.MeterGaugeChart",
					"org.primefaces.component.chart.MeterGaugeChartRenderer");
			meterGaugeChart.setTitle(displayConnectorName);

			// Add the corresponding model
			meterGaugeChartModel = new MeterGaugeChartModel();
			meterGaugeChart.setValue(meterGaugeChartModel);

			component.getChildren().clear();
			component.getChildren().add(meterGaugeChart);
		}

		try {
			final Object displayObj = this.projectService.getDisplay(this.projectName, this.activeView.getName(), displayConnectorName);

			final Set<String> keys = (Set<String>) new Mirror().on(displayObj).invoke().method("getKeys").withoutArgs();
			if (keys.isEmpty()) {
				meterGaugeChartModel.setIntervals(Arrays.asList((Number) 50));
				meterGaugeChartModel.setValue(0);
			} else {
				final List<Number> intervals = (List<Number>) new Mirror().on(displayObj).invoke().method("getIntervals").withArgs(keys.iterator().next());
				final Number value = (Number) new Mirror().on(displayObj).invoke().method("getValue").withArgs(keys.iterator().next());

				if (intervals.isEmpty()) {
					meterGaugeChartModel.setIntervals(Arrays.asList((Number) 50));
					meterGaugeChartModel.setValue(0);
				} else {
					meterGaugeChartModel.setIntervals(intervals);
					meterGaugeChart.setShadow(true);
					meterGaugeChart.setSeriesColors("66cc66, 93b75f, E7E658, cc6666");
					final Number maxInterval = intervals.get(intervals.size() - 1);
					if (value.doubleValue() <= maxInterval.doubleValue()) {
						meterGaugeChartModel.setValue(value);
					} else {
						meterGaugeChartModel.setValue(maxInterval);
					}
				}
			}
		} catch (final DisplayNotFoundException ex) {
			CurrentCockpitBean.LOG.warn("Display not found.", ex);
			this.updateDisplayWithDefaultContent(component);
		} catch (final MirrorException ex) {
			CurrentCockpitBean.LOG.warn("Reflection exception.", ex);
			this.updateDisplayWithDefaultContent(component);
		} catch (final InvalidAnalysisStateException ex) {
			CurrentCockpitBean.LOG.info("Project is in invalid state.", ex);
			this.updateDisplayWithDefaultContent(component);
		}
	}

	@SuppressWarnings("unchecked")
	private void updateXYPlotDisplay(final UIComponent component, final String displayConnectorName) {
		final boolean isAlreadyCorrectlyInitialized = (component.getChildCount() == 1) && (component.getChildren().get(0) instanceof LineChart);

		final LineChartSeries lineChartSeries;
		final SelectCheckboxMenu selectCheckboxMenu;

		// Avoid unnecessary component creation. Use the existing components if they are already existing and from the correct type.
		if (isAlreadyCorrectlyInitialized) {
			final LineChart lineChart = (LineChart) component.getChildren().get(0);
			lineChartSeries = (LineChartSeries) ((CartesianChartModel) lineChart.getValue()).getSeries().get(0);
			selectCheckboxMenu = (SelectCheckboxMenu) component.getChildren().get(1);
		} else {
			final FacesContext facesContext = FacesContext.getCurrentInstance();
			final Application application = facesContext.getApplication();

			// Create the Primefaces chart component
			final LineChart lineChart = (LineChart) application.createComponent(facesContext, "org.primefaces.component.chart.LineChart",
					"org.primefaces.component.chart.LineChartRenderer");
			lineChart.setTitle(displayConnectorName);

			// Add the corresponding model
			final CartesianChartModel linearModel = new CartesianChartModel();
			lineChartSeries = new LineChartSeries();
			lineChartSeries.setLabel("Series");

			linearModel.addSeries(lineChartSeries);
			lineChart.setValue(linearModel);

			// Create the Primefaces drop box
			selectCheckboxMenu = (SelectCheckboxMenu) application.createComponent(facesContext, "org.primefaces.component.SelectCheckboxMenu",
					"org.primefaces.component.SelectCheckboxMenuRenderer");

			component.getChildren().clear();
			component.getChildren().add(lineChart);
			component.getChildren().add(selectCheckboxMenu);
		}

		try {
			final Object displayObj = this.projectService.getDisplay(this.projectName, this.activeView.getName(), displayConnectorName);
			final Set<String> keys = (Set<String>) new Mirror().on(displayObj).invoke().method("getKeys").withoutArgs();

			if (keys.isEmpty()) {
				lineChartSeries.set(0, 0);
			} else {
				final Map<Object, Number> entries = (Map<Object, Number>) new Mirror().on(displayObj).invoke().method("getEntries").withArgs(keys.iterator().next());

				if (entries.isEmpty()) {
					lineChartSeries.set(0, 0);
				} else {
					lineChartSeries.setData(entries);
				}
			}
		} catch (final DisplayNotFoundException ex) {
			CurrentCockpitBean.LOG.warn("Display not found.", ex);
			this.updateDisplayWithDefaultContent(component);
		} catch (final MirrorException ex) {
			CurrentCockpitBean.LOG.warn("Reflection exception.", ex);
			this.updateDisplayWithDefaultContent(component);
		} catch (final InvalidAnalysisStateException ex) {
			CurrentCockpitBean.LOG.info("Project is in invalid state.", ex);
			this.updateDisplayWithDefaultContent(component);
		}

	}

}
